########################################################################
# Return the calculated position of Stars
########################################################################
import logging
import os
from datetime import datetime
import string

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

import star_chart_spherical_projection

## Logging set up
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
stream_handler = logging.StreamHandler()
logger.addHandler(stream_handler)

def final_position(included_stars=[], 
                    year_since_2000=0,
                    is_precession=True,
                    added_stars=[],
                    only_added_stars=False,
                    declination_min=None,
                    declination_max=None,
                    save_to_csv=None):
    # return the final position of the stars as a dictionary

    star_chart_spherical_projection.errorHandling(isPlotFunction=False,
                                                included_stars=included_stars,
                                                year_since_2000=year_since_2000,
                                                is_precession=is_precession,
                                                added_stars=added_stars,
                                                only_added_stars=only_added_stars,
                                                declination_min=declination_min,
                                                declination_max=declination_max,
                                                save_to_csv=save_to_csv)

    if not only_added_stars:
        # show all stars
        included_stars = [string.capwords(x) for x in included_stars]
        listOfStars = star_chart_spherical_projection._get_stars(included_stars)
        for star_object in added_stars:
            star_row = [star_object.star_name,
                        star_object.ra,
                        star_object.dec,
                        star_object.pm_speed,
                        star_object.pm_angle,
                        star_object.magnitude]
            listOfStars.append(star_row)
    else:
        # show only the user's added stars
        listOfStars = []
        for star_object in added_stars:
            star_row = [star_object.star_name,
                        star_object.ra,
                        star_object.dec,
                        star_object.pm_speed,
                        star_object.pm_angle,
                        star_object.magnitude]
            listOfStars.append(star_row)
    
    # Set declination min values when using the _generate_stereographic_projection() to capture all stars if not set
    declination_min = -90
    declination_max = 90

    _, _, _, finalPositionOfStarsDict = star_chart_spherical_projection._generate_stereographic_projection(starList=listOfStars, 
                                                                                                        pole="North", 
                                                                                                        declination_min=declination_min,
                                                                                                        year_since_2000=year_since_2000,
                                                                                                        is_precession=is_precession,
                                                                                                        max_magnitude=None,
                                                                                                        declination_max=declination_max)
    # Generate a .csv file with final positions of stars
    if save_to_csv is not None:
        header_options = ["Common Name", "Right Ascension (HH.MM.SS)", "Declination (DD.SS)"]
        star_chart_list = []
        for star_name, star_position in finalPositionOfStarsDict.items():
            star_chart_list.append([star_name, star_position["RA"], star_position["Declination"]])
        df = pd.DataFrame(star_chart_list, columns=header_options)
        df = df.sort_values(by=["Common Name"])
        df.to_csv(save_to_csv, header=header_options, index=False)

    return finalPositionOfStarsDict

def position_over_time(star=None,
                        added_star=None,
                        start_year_since_2000=None,
                        end_year_since_2000=None,
                        increment=5,
                        is_precession=True,
                        save_to_csv=None):

    if star is not None:
        star_csv_file = os.path.join(os.path.dirname(__file__), 'data', 'stars_with_data.csv')  # get file's directory, up one level, /data/4_all_stars_data.csv
        star_dataframe = pd.read_csv(star_csv_file)
        star_data = star_dataframe.loc[star_dataframe["Common Name"] == star].values.flatten().tolist()
        star_name, star_ra, star_declination, star_mag, star_pm_speed, star_pm_angle, star_pm_ra, star_pm_dec, star_alt_names, star_url = star_data
    if added_star is not None:
        star_name = added_star.star_name
        star_ra = added_star.ra
        star_declination = added_star.dec
        star_pm_speed = added_star.pm_speed
        star_pm_angle = added_star.pm_angle
        star_mag = added_star.magnitude

    years_to_calculate = np.arange(start_year_since_2000, end_year_since_2000+1, increment).tolist()
    position_over_time = {}
    for year in years_to_calculate:
        star_row = [[star_name, star_ra, star_declination, star_pm_speed, star_pm_angle, star_mag]]
        _, star_radians, _, star_dict = star_chart_spherical_projection._generate_stereographic_projection(starList=star_row, 
                                                                                            year_since_2000=year,
                                                                                            is_precession=is_precession,
                                                                                            pole="North",
                                                                                            declination_min=-90,
                                                                                            declination_max=90)
        
        position_over_time[year+2000] = {"RA (radians)": star_radians[0], 
                                    "RA (hours)" : star_dict[star_name]["RA"], 
                                    "Dec (degrees)" : star_dict[star_name]["Declination"]}
    
    # Generate a .csv file with final positions of the star
    if save_to_csv is not None:
        header_options = ["Year", "Declination (DD.SS)", "Right Ascension (HH.MM.SS)", "Right Ascension (radians)"]
        star_chart_list = []
        for year, star_position in position_over_time.items():
            star_chart_list.append([year, star_position["Dec (degrees)"], star_position["RA (hours)"], star_position["RA (radians)"]])
        df = pd.DataFrame(star_chart_list, columns=header_options)
        df = df.sort_values(by=["Year"])
        df.to_csv(save_to_csv, header=header_options, index=False)

    return position_over_time

def plot_position(star=None, 
                added_star=None,
                start_year_since_2000=None,
                end_year_since_2000=None,
                increment=10,
                is_precession=True,
                dec_ra="D",
                show_plot=True,
                display_year_marker=True,
                fig_plot_title=None,
                fig_plot_color="C0",
                figsize_n=12,
                figsize_dpi=100,
                save_plot_name=None):

    position_over_time_dict = position_over_time(star=star,
                                                added_star=added_star,
                                                start_year_since_2000=start_year_since_2000,
                                                end_year_since_2000=end_year_since_2000,
                                                increment=increment,
                                                is_precession=is_precession)

    if star is not None:
        star_csv_file = os.path.join(os.path.dirname(__file__), 'data', 'stars_with_data.csv')  # get file's directory, up one level, /data/4_all_stars_data.csv
        star_dataframe = pd.read_csv(star_csv_file)
        star_data = star_dataframe.loc[star_dataframe["Common Name"] == star].values.flatten().tolist()
        star_name = star_data[0]
    if added_star is not None:
        star_name = added_star.star_name
    
    fig = plt.figure(figsize=(figsize_n,figsize_n), dpi=figsize_dpi)
    ax = fig.subplots()

    year_lst = []
    dec_lst = []
    ra_radians_lst = []
    for year, position_dict in position_over_time_dict.items():
        year_lst.append(year)
        dec_lst.append(position_dict["Dec (degrees)"])
        ra_radians_lst.append(position_dict["RA (radians)"])

    if dec_ra == "D":
        plot_y = dec_lst
        title = "Declination"
        y_label = "Declination (°)"
    if dec_ra == "R":
        plot_y = ra_radians_lst
        title = "Right Ascension"
        y_label = "Right Ascension (Radians)"

    if is_precession:
        precession_label = "(With Precession)"
    else:
        precession_label = "(Without Precession)"

    # prevent the axis from populate more than 70 elements (to prevent overlapping)
    x_increment = increment
    x_plot_len = np.arange(year_lst[0], year_lst[-1]+1, x_increment)
    while len(x_plot_len) > 71:
        x_increment *= 5
        x_plot_len = np.arange(year_lst[0], year_lst[-1]+1, x_increment)
    y_increment = len(plot_y)
    while y_increment > 20:
        y_increment = 20

    if year_lst[0] >= -2000: startYear_bce_ce = f"{year_lst[0]} C.E" # positive years for C.E
    if year_lst[0] < -2000: startYear_bce_ce = f"{abs(year_lst[0])} B.C.E" # negative years for B.C.E
    if year_lst[-1] >= -2000: endYear_bce_ce = f"{year_lst[-1]} C.E" # positive years for C.E
    if year_lst[-1] < -2000: endYear_bce_ce = f"{abs(year_lst[-1])} B.C.E" # negative years for B.C.E
    
    plt.title(f"{star_name}'s {title} {precession_label} from {startYear_bce_ce} to {endYear_bce_ce}, Every {increment} Years")
    plt.plot(year_lst, plot_y)
    plt.xlabel("Year")
    plt.ylabel(y_label)
    
    ax.set_xticks(np.arange(year_lst[0], year_lst[-1]+1, x_increment))
    ax.set_yticks(np.linspace(min(plot_y), max(plot_y), y_increment))

    if display_year_marker:
        current_year = datetime.now().year
        plt.axvline(current_year, linewidth=0.5, color="black", linestyle="dashed")
    plt.xticks(rotation=90)

    if show_plot:
        plt.show()

    if save_plot_name is not None:
        fig.savefig(save_plot_name, dpi=fig.dpi)

def predict_pole_star(year_since_2000=0, pole="North", max_magnitude=None):
    # Find the next North/South Pole Star

    star_chart_spherical_projection.errorHandlingPredictPoleStar(year_since_2000=year_since_2000, pole=pole)
    pole = pole.title()

    final_position_builtin_stars = star_chart_spherical_projection.final_position(year_since_2000=year_since_2000,
                                                                                is_precession=True)
    # Set the pole declination based on either North/South
    if pole == "North":
        pole_declination = 90
    if pole == "South":
        pole_declination = -90
    
    # Collect data to check magnitude of each star
    star_csv_file = os.path.join(os.path.dirname(__file__), 'data', 'stars_with_data.csv')  # get file's directory, up one level, /data/4_all_stars_data.csv
    star_dataframe = pd.read_csv(star_csv_file)

    # Set max_magnitude to max magnitude in star data if set to None
    if max_magnitude is None:
        max_magnitude = star_dataframe["Magnitude (V, Visual)"].max()

    # Find the closest star in built-in stars
    closest_pole_star = None
    closest_pole_declination = None
    for star, star_data in final_position_builtin_stars.items():
        star_mag = star_dataframe.loc[star_dataframe["Common Name"]==star]["Magnitude (V, Visual)"].values[0]
        if star_mag <= max_magnitude: # filter out stars with magnitudes larger than filter
            if closest_pole_star is None:
                closest_pole_star = star
                closest_pole_declination = star_data["Declination"]
            else:
                if abs(float(star_data["Declination"]) - pole_declination) < abs(closest_pole_declination - pole_declination):
                    closest_pole_star = star
                    closest_pole_declination = star_data["Declination"]
        
    return closest_pole_star



########################################################################
# ERROR CATCHES AND LOGGING
########################################################################
import csv
import numpy as np
import os
import pandas as pd

import star_chart_spherical_projection

def errorHandling(isPlotFunction=None,
                included_stars=None,
                pole=None, 
                year_since_2000=None,
                is_precession=None,
                declination_max=None,
                declination_min=None,
                max_magnitude=None,
                added_stars=None,
                only_added_stars=None,
                display_labels=None,
                display_dec=None,
                increment=None,
                show_plot=None,
                fig_plot_title=None,
                fig_plot_color=None,
                figsize_n=None,
                figsize_dpi=None,
                save_plot_name=None,
                save_to_csv=None):
    # Error Handling for Variables shared between final_position() and plot_stereographic_projection() (defined by isPlotFunction)

    # Ensure that star list is a list
    if type(included_stars) != list:
        raise ValueError(f"[included_stars]: Must be a list, current type = '{type(included_stars)}'")

    ## Check that user list has stars that are found in current list
    if len(included_stars) != 0:
        all_common_names = [name[0] for name in star_chart_spherical_projection._get_stars()]
        for star_given in included_stars:
            if star_given.lower() not in [x.lower() for x in all_common_names]: # convert all stars to lower case:
                raise ValueError(f"[included_stars]: '{star_given}' not a star in current list of stars, please select one of the following: {all_common_names}")

    # Ensure that declination ranges are set and within within ranges
    if declination_min is not None:
        if type(declination_min) != int and type(declination_min) != float:
            raise ValueError(f"[declination_min]: Must be a int or float, current type = '{type(declination_min)}'")
        if declination_min not in np.arange(-89, 90): # if defined, but not in range
            raise ValueError(f"[declination_min]: Minimum declination must lie between -90 and +90 (-89 to 89) [recommended by default: north=-30, south=30], current minimum = '{declination_min}'")

    # Ensure if a year is selected it is a float or int, set by default to 0 (the year = 2000)
    if type(year_since_2000) != int and type(year_since_2000) != float:
        raise ValueError(f"[year_since_2000]: Must be a int or float, current type = '{type(year_since_2000)}'")

    # Ensure that precession options are booleans ["True", "False"]
    if type(is_precession) != bool:
        raise ValueError(f"[is_precession]: Must be a bool, current type = '{type(is_precession)}'")

    if type(added_stars) != list:
        raise ValueError(f"[added_stars]: Must be a list, current type = '{type(added_stars)}'")
    for user_star in added_stars:
        if type(user_star) != star_chart_spherical_projection.add_new_star:
            raise ValueError(f"[added_stars]: {type(user_star)} is not a valid new star object (see: star_chart_spherical_projection.add_new_star)")

    if type(only_added_stars) != bool:
        raise ValueError(f"[only_added_stars]: Must be a bool, current type = '{type(only_added_stars)}'")

    # Error Handling for final_position() function
    if not isPlotFunction:
        # Ensure that declination ranges are set and within within ranges
        if declination_max is not None:
            if type(declination_max) != int and type(declination_max) != float:
                raise ValueError(f"[declination_max]: Must be a int or float, current type = '{type(declination_max)}'")
            if declination_max not in np.arange(-89, 90): # if defined, but not in range
                raise ValueError(f"[declination_max]: Maximum declination must lie between -90 and +90 (-89 to 89) [recommended by default: north=-30, south=30], current minimum = '{declination_max}'")
        if save_to_csv is not None:
            if type(save_to_csv) != str:
                raise ValueError(f"[save_to_csv]: Must be a str, current type = '{type(save_to_csv)}'")
            if not save_to_csv.lower().endswith(".csv"):
                raise ValueError(f"[save_to_csv]: Extension must be a .csv file, current extension = '{save_to_csv.split('.')[1]}'")

    # Error Handling for plot_stereographic_projection() function
    if isPlotFunction:
        # Ensure that Pole selected are within options
        if type(pole) != str:
            raise ValueError(f"[pole]: Must be a str, current type = '{type(pole)}'")
        else:
            if pole not in ["North", "South"]:
                raise ValueError(f"[pole]: Pole options are ['North', 'South'], current option = '{pole}'")

        # Ensure that max_magnitude options is a float, set by default to None
        if max_magnitude is not None:
            if type(max_magnitude) != int and type(max_magnitude) != float:
                raise ValueError(f"[max_magnitude]: Must be a int or float, current type = '{type(max_magnitude)}'")

        # Ensure that the display options for star names and declination numbers are booleans ["True", "False"]
        if type(display_labels) != bool:
            raise ValueError(f"[display_labels]: Must be a bool, current type = '{type(display_labels)}'")

        if type(display_dec) != bool:
            raise ValueError(f"[display_dec]: Must be a bool, current type = '{type(display_dec)}'")

        # Ensure that increment options are 1, 5, 10
        if type(increment) != int:
            raise ValueError(f"[increment]: Must be a int, current type = '{type(increment)}'")
        if increment not in [1, 5, 10]:
            raise ValueError(f"[increment]: Must be one of the options [1, 5, 10], current value = '{increment}'")

        # Ensure that the only options for show_plot are booleans ["True", "False"]
        if type(show_plot) != bool:
            raise ValueError(f"[show_plot]: Must be a bool, current type = '{type(show_plot)}'")

        # Ensure that the color given is a string (matplotlib has error checking for invalid color options)
        if type(fig_plot_color) != str:
            raise ValueError(f"[fig_plot_color]: Must be a string, current type = '{type(fig_plot_color)}'")

        # Ensure that the user defined title of the plot is a string
        if fig_plot_title is not None and type(fig_plot_title) != str:
            raise ValueError(f"[fig_plot_title]: Must be a string, current type = '{type(fig_plot_title)}'")

        # Ensure that figure options are integers or floats
        if type(figsize_n) != int and type(figsize_n) != float:
            raise ValueError(f"[figsize_n]: Must be a int or float, current type = '{type(figsize_n)}'")
        if type(figsize_dpi) != int and type(figsize_dpi) != float:
            raise ValueError(f"[figsize_dpi]: Must be a int or float, current type = '{type(figsize_dpi)}'")

        # Ensure that the user defined figure saved name is a string
        if save_plot_name is not None and type(save_plot_name) != str:
            raise ValueError(f"[save_plot_name]: Must be a string, current type = '{type(save_plot_name)}'")

def errorHandlingStarClass(star_name=None,
                        ra=None,
                        dec=None,
                        pm_speed=None,
                        pm_angle=None,
                        pm_speed_ra=None,
                        pm_speed_dec=None,
                        magnitude=None):
    
    if star_name is None:
        raise ValueError("[star_name]: star_name is required")
    else:
        if type(star_name) != str:
            raise ValueError(f"[star_name]: Must be a str, current type = '{type(star_name)}'")

    if ra is None:
        raise ValueError("[ra]: Right Ascension is required")
    else:
        if type(ra) != str:
            raise ValueError(f"[ra]: Must be a str, current type = '{type(ra)}'")
        ra_items = ra.split('.')
        if len(ra_items) != 3:
            raise ValueError(f"[ra]: Right Ascension must be three parts '[HH, MM, SS]' (Hours, Minutes, Seconds), currently  = '{ra_items}'")
        for ra_time in ra_items:
            if not ra_time.isdigit():
                raise ValueError(f"[ra]: Each part of the Right Ascension must be an integer, '{ra_time}' current type = {type(ra_time)}")

    if dec is None:
        raise ValueError("[dec]: Declination is required")
    else:
        if type(dec) != int and type(dec) != float:
            raise ValueError(f"[dec]: Must be a int or float, current type = '{type(dec)}'")

    if pm_speed is not None:
        if type(pm_speed) != int and type(pm_speed) != float:
            raise ValueError(f"[pm_speed]: Must be a int or float, current type = '{type(pm_speed)}'")

    if pm_angle is not None:
        if type(pm_angle) != int and type(pm_angle) != float:
            raise ValueError(f"[pm_angle]: Must be a int or float, current type = '{type(pm_angle)}'")

    if pm_speed_ra is not None:
        if type(pm_speed_ra) != int and type(pm_speed_ra) != float:
            raise ValueError(f"[pm_speed_ra]: Must be a int or float, current type = '{type(pm_speed_ra)}'")

    if pm_speed_dec is not None:
        if type(pm_speed_dec) != int and type(pm_speed_dec) != float:
            raise ValueError(f"[pm_speed_dec]: Must be a int or float, current type = '{type(pm_speed_dec)}'")

    # Verify at least one pair is set
    if pm_speed_ra is None and pm_speed_dec is None:
        # Neither pairs are set
        if pm_speed is None and pm_angle is None:
            raise ValueError("Either pm_speed_ra/pm_speed_dec or pm_speed/pm_angle is required")

    # Verify when only one value is set, make sure to set up its pair
    if pm_speed_ra is not None:
        if pm_speed_dec is None and pm_speed is None and pm_angle is None:
            raise ValueError("[pm_speed_dec]: With pm_speed_ra set, pm_speed_dec is required")

    if pm_speed_dec is not None:
        if pm_speed_ra is None and pm_speed is None and pm_angle is None:
            raise ValueError("[pm_speed_ra]: With pm_speed_dec set, pm_speed_ra is required")
    if pm_speed is not None:
        if pm_speed_ra is None and pm_speed_dec is None and pm_angle is None:
            raise ValueError("[pm_angle]: With pm_speed set, pm_angle is required")
    if pm_angle is not None:
        if pm_speed_ra is None and pm_speed_dec is None and pm_speed is None:
            raise ValueError("[pm_speed]: With pm_angle set, pm_speed is required")

    # Remove the invalid extra options when pm_speed/pm_angle should be the only pair set
    if pm_speed is not None and pm_angle is not None:
        if pm_speed_ra is not None and pm_speed_dec is not None:
            raise ValueError("Either pm_speed_ra/pm_speed_dec or pm_speed/pm_angle is required, not both")
        if pm_speed_ra is not None and pm_speed_dec is None:
            raise ValueError("[pm_speed_ra]: With pm_speed/pm_angle set, pm_speed_ra should be None")
        if pm_speed_ra is None and pm_speed_dec is not None:
            raise ValueError("[pm_speed_dec]: With pm_speed/pm_angle set, pm_speed_dec should be None")

    # Remove the invalid extra options when pm_speed_ra/pm_speed_dec should be the only pair set
    if pm_speed_ra is not None and pm_speed_dec is not None:
        if pm_speed is None and pm_angle is not None:
            raise ValueError("[pm_angle]: With pm_speed_ra/pm_speed_dec set, pm_angle should be None")
        if pm_speed is not None and pm_angle is None:
            raise ValueError("[pm_speed]: With pm_speed_ra/pm_speed_dec set, pm_speed should be None")

    # Verify the non-None values are the correct pairs: pm_speed_ra/pm_speed_dec or pm_speed/pm_angle
    if pm_speed is None and pm_speed_dec is None:
        if pm_speed_ra is not None and pm_angle is not None:
            raise ValueError("Should be a pair of pm_speed_ra/pm_speed_dec or pm_speed/pm_angle, not pm_angle/pm_speed_ra")
    if pm_speed is None and pm_speed_ra is None:
        if pm_angle is not None and pm_speed_dec is not None:
            raise ValueError("Should be a pair of pm_speed_ra/pm_speed_dec or pm_speed/pm_angle, not pm_angle/pm_speed_dec")
    if pm_angle is None and pm_speed_ra is None:
        if pm_speed is not None and pm_speed_dec is not None:
            raise ValueError("Should be a pair of pm_speed_ra/pm_speed_dec or pm_speed/pm_angle, not pm_speed/pm_speed_dec")
    if pm_angle is None and pm_speed_dec is None:
        if pm_speed is not None and pm_speed_ra is not None:
            raise ValueError("Should be a pair of pm_speed_ra/pm_speed_dec or pm_speed/pm_angle, not pm_speed/pm_speed_ra")

    if magnitude is None:
        raise ValueError("[magnitude]: magnitude is required")
    else:
        if type(magnitude) != int and type(magnitude) != float:
            raise ValueError(f"[magnitude]: Must be a int or float, current type = '{type(magnitude)}'")

def errorHandlingPredictPoleStar(year_since_2000=None, pole=None):
    # Error Handling for predict_pole_star()
    if year_since_2000 is None:
        raise ValueError("[year_since_2000]: year_since_2000 is required")
    else:
        if type(year_since_2000) != int and type(year_since_2000) != float:
            raise ValueError(f"[year_since_2000]: Must be a int or float, current type = '{type(year_since_2000)}'")
    
    if pole is not None:
        if type(pole) != str:
            raise ValueError(f"[pole]: Must be a str, current type = '{type(pole)}'")
        else:
            if pole.title() not in ["North", "South"]:
                raise ValueError(f"[pole]: Must be a 'North' or 'South', currently = '{pole}'")



########################################################################
# Generate a star chart centered on the poles:
#       Northern Hemisphere = 90°
#       Southern Hemisphere = -90°
########################################################################
import csv
import logging
import math
import matplotlib.pyplot as plt
import numpy as np
import os
import pandas as pd

import star_chart_spherical_projection

## Logging set up
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
stream_handler = logging.StreamHandler()
logger.addHandler(stream_handler)

## Constants
northern_declination_min = -30
northern_declination_max = 90
southern_declination_min = 30
southern_declination_max = -90

# Start Year (JP2000)
j2000 = 2000 # start year of the star catalogue (jan 1 2000 via IAU)

def _get_stars(selectStars=[]):
    # selectStars only returns a subset of all the stars saved, empty will return all in the star_data.csv file
    # stars: ["name", "RA: HH.MM.SS", Declination DD.SS, Proper Motion Speed (mas/yr), Proper Motion Angle (DD.SS), Magnitude (V, Visual)]
    star_data_list = []
    star_csv_file = os.path.join(os.path.dirname(__file__), 'data', 'stars_with_data.csv')  # get file's directory, up one level, /data/4_all_stars_data.csv
    star_dataframe = pd.read_csv(star_csv_file)
    for index, row in star_dataframe.iterrows():
        if len(selectStars) > 0: # get only a subset of all stars
            if row["Common Name"] in selectStars:
                star_data_list.append(row.tolist())
        else:
            star_data_list.append(row.tolist())
    return star_data_list

def _ra_to_radians(star_list):
    # convert RA to radians
    for star in star_list:
        # convert RA from hours to degrees
        ra_in_hr = star[1]
        
        ra_hr, ra_min, ra_sec = ra_in_hr.split('.')    
        ra_hr = int(ra_hr)
        ra_min = int(ra_min)
        prec_zero = len(ra_sec)-len(ra_sec.lstrip('0')) # count number of preceding zeroes (for example, 0.00868)
        
        # if seconds have greater than 2 degrees of accuracy (for example: 07.46.519615)
        len_degrees = 0
        if prec_zero >= 30: # account for measuring the length of a number in scientifiic notation
            len_degrees = -1
        len_degrees += len(ra_sec)-prec_zero
        if len_degrees < 0: len_degrees = 0
        second_degrees = 10**(len_degrees + prec_zero) 
        ra_sec = int(ra_sec)
        ra_sec /= second_degrees # divide seconds by a multiple (for example, to convert: 519615 to 51.9615)
        
        # convert minutes and seconds to decimals
        ra_min /= 60
        ra_sec /= 3600
        ra_total = ra_hr + ra_min + ra_sec

        # convert RA from degrees to radians
        ra_in_degrees = ra_total * 15
        ra_in_radians = np.deg2rad(ra_in_degrees)
        star[1] = ra_in_radians
 
    return star_list

def _radians_to_ra(ra_in_radians):
    # change star in radians to RA in hours
    ra_in_degree = np.rad2deg(ra_in_radians)

    if ra_in_degree > 360 or ra_in_degree < 0: # lock degrees between 0 and 360, if negative, re-write as a positive degree
        ra_in_degree %= 360
    
    hours = int(ra_in_degree / 15)
    minutes = int(((ra_in_degree / 15) - hours) * 60) # measured in minutes
    seconds = round(((((ra_in_degree / 15) - hours) * 60) - minutes) * 60, 10) # measured in seconds
    
    precision = len(str(seconds).split(".")[1])
    if precision > 5:
        seconds = f"{seconds:.10f}".rstrip("0") #str(float(f"{seconds:.{precision}}"))# # account for scientific notation (for example, convert 8.68e-05 to 0.0000868)
    else:
        seconds = str(seconds)
    prec_seconds = seconds.split(".")[1]
    if prec_seconds == "": prec_seconds = 0

    # count preceding zeroes
    prec_zero = 0 
    if int(prec_seconds) != 0:
        prec_zero = len(prec_seconds)-len(prec_seconds.lstrip('0')) # count number of preceding zeroes (for example, 0.00868)
    
    # handle edge case to convert "0870314" to "870314" to "8.70314" when value is a long decimal
    is_less_ten = False 
    if float(seconds) < 10: is_less_ten = True
    if not float(seconds).is_integer(): # if float/decimal (for example, 54.47676 becomes 5447676, but 9.0 stays at 9)
        # convert from decimal to whole number while maintaining precession
        seconds = int(str(seconds).replace(".", ""))
    else:
        # convert float of integer to integer (9.0 = 9)
        seconds = int(float(seconds))
    
    # handle edge case if seconds == 60, increment minutes by one
    if seconds == 60:
        minutes += 1
        seconds = 0

    # RA in hours 'HH.MM.SS'
    if hours < 10: hours = '0' + str(hours) # convert 6 to 06
    if minutes < 10: minutes = '0' + str(minutes) # convert 6 to 06
    if (seconds < 10 and seconds > 0) or is_less_ten:
        seconds = ('0'*prec_zero) + str(seconds) # dynamically add zeroes to front of string 
    if seconds == 0:
        seconds = "00" # add trailing zero when exactly 0
    if len(str(seconds)) == 1: # add trailing zero for 1 -> 10
        seconds += "0"

    ra_in_hours = f"{hours}.{minutes}.{seconds}"
    return ra_in_hours

def _ra_dec_via_pm(years_since_2000, star_ra, star_dec, star_pm_speed, star_pm_angle):
    # Calculate the RA and Declination of a star based on changes due to Proper Motion
    # returns calculated RA and Declination

    logger.debug(f"Proper Motion for {years_since_2000} Years")
    logger.debug(f"Date {years_since_2000}, RA = {star_ra}, Dec = {star_dec}, PM Speed = {star_pm_speed}, PM Angle = {star_pm_angle}")

    star_pm_speed_degrees = 0.00000027777776630942 * star_pm_speed # convert mas/yr to degrees/yr
    star_pm_speed_radians = np.deg2rad(star_pm_speed_degrees) # radians/yr
    star_movement_radians_per_year = star_pm_speed_radians * years_since_2000
    logger.debug(f"Movement Over Time = {star_movement_radians_per_year} (rad), {star_movement_radians_per_year} (deg)")

    ra_x_difference_component = star_movement_radians_per_year * math.cos(np.deg2rad(star_pm_angle))
    dec_y_difference_component = star_movement_radians_per_year * math.sin(np.deg2rad(star_pm_angle))
    logger.debug(f"(RA)  x Difference = {ra_x_difference_component} (rad) = {np.rad2deg(ra_x_difference_component)} degrees")
    logger.debug(f"(DEC) y Difference = {dec_y_difference_component} (rad) = {np.rad2deg(dec_y_difference_component)} degrees")

    star_adjusted_ra = star_ra + ra_x_difference_component # in radians with proper motion (potentially  will be flipped 180 based on new declination)
    star_adjusted_declination = star_dec + np.rad2deg(dec_y_difference_component) # in degrees new with proper motion

    dec_x = star_adjusted_declination
    # remap within -90 and 90 for positive declinations
    if star_adjusted_declination > 0: # Positive declinations
        dec_x = star_adjusted_declination % 360
        # map from 0 to 90 (positive declinations)
        if dec_x > 90 and dec_x <= 180: 
            dec_x = 90 + (90 - dec_x)
        # map from 0 to -90 (negative declinations)
        if dec_x <= 270 and dec_x > 180:
            dec_x = 180 - dec_x
        if dec_x < 360 and dec_x > 270:
            dec_x = -90 + (dec_x - 270)
    if star_adjusted_declination < -0: # Negative declinations
        dec_x = star_adjusted_declination % -360
        # map from 0 to -90 (negative declinations)
        if dec_x < -90 and dec_x >= -180: 
            dec_x = -90 - (90 + dec_x)
        # map from 0 to 90 (positive declinations)
        if dec_x >= -270 and dec_x <= -180:
            dec_x = 180 + dec_x
            dec_x = abs(dec_x)
        if dec_x > -360 and dec_x < -270:
            dec_x = 90 + (dec_x + 270)
    logger.debug(f"New mapped dec = {dec_x}")
    logger.debug(f"Original Dec = {star_dec}, New Dec = {star_adjusted_declination}")

    # flip over RA by rotating 180
    is_flipped_across_pole = False
    star_over_ninety_ra = star_adjusted_declination
    if star_over_ninety_ra >= 0: # positive declinations
        while (star_over_ninety_ra > 90):
            star_over_ninety_ra -= 90
            is_flipped_across_pole = not is_flipped_across_pole # flip across the center by 180
    if star_over_ninety_ra < 0: # negative declinations
        while (star_over_ninety_ra < -90):
            star_over_ninety_ra += 90
            is_flipped_across_pole = not is_flipped_across_pole # flip across the center by 180
    logger.debug(f"Original RA = {np.rad2deg(star_ra)}, New RA = {np.rad2deg(star_adjusted_ra)}, Flipped = {is_flipped_across_pole} (if true +180?)")

    # If declination goes over ninety, flip over by 180
    if is_flipped_across_pole:
        star_adjusted_ra = star_adjusted_ra + np.deg2rad(180)

    star_adjusted_declination = dec_x
    logger.debug(f"Final RA: {np.rad2deg(star_adjusted_ra)} degrees")
    logger.debug(f"Final Dec: {star_adjusted_declination} degrees")
    return star_adjusted_ra, star_adjusted_declination

def calculatePositionOfPolePrecession(years_since_2000, original_declination, original_ra):
    # Calculate change in the position of the pole due to precession
    obliquity_for_YYYY = 23.439167
    logger.debug(f"Years Since 2000 = {years_since_2000}")

    # Rate of change of right ascension and declination of a star due to precession
    declination_change_arcseconds_per_year = (19.9 * math.cos(original_ra)) * years_since_2000
    ra_change_arcseconds_per_year = (46.1 * (19.9 * math.sin(original_ra) * math.tan(np.deg2rad(original_declination)))) * years_since_2000

    change_in_declination = declination_change_arcseconds_per_year/3600 # degrees
    change_in_ra = np.deg2rad(ra_change_arcseconds_per_year/3600) # degrees to radians

    final_declination_due_to_precession = original_declination + change_in_declination
    final_ra_due_to_precession = original_ra + change_in_ra

    logger.debug(f"Dec: {original_declination} + {change_in_declination} = {final_declination_due_to_precession}")
    logger.debug(f"RA:  {original_ra} + {change_in_ra} = {final_ra_due_to_precession}")
    return final_ra_due_to_precession, final_declination_due_to_precession

def vondrakDreamalligator(star_name, star_ra_rad, star_dec_rad, year_to_calculate):
    # Modified code from github.com/dreamalligator/vondrak to calculate precession
    def position_matrix(ra=None, dec=None):
        x = math.cos(dec) * math.cos(ra)
        y = math.cos(dec) * math.sin(ra)
        z = math.sin(dec)
        return np.array([[x], [y], [z]])

    def compute_star(compute_ra, compute_dec):
        return position_matrix(ra=compute_ra, dec=compute_dec)

    precession_matrix = star_chart_spherical_projection.ltp_pbmat(year_to_calculate) # Precession matrix for the given year
    p_1 = compute_star(star_ra_rad, star_dec_rad) # compute star's position matrix for given year
    p_1 = star_chart_spherical_projection.pdp(precession_matrix, p_1) # apply precession matrix for given year
    (ra_as_rad, dec_as_rad) = star_chart_spherical_projection.ra_dec(p_1)

    return dec_as_rad, ra_as_rad # new declination and right ascension

def _precession_vondrak(star_name, star_ra, star_dec, year_YYYY_since_2000):
    # Temporary fix for vondrak plugin (will only find a smaller subsections of the stars)
    logger.debug("INCLUDING PRECESSION VIA VONDRAK")
    vondrak_dec, vondrak_ra = vondrakDreamalligator(star_name, star_ra, np.deg2rad(star_dec), 2000 + year_YYYY_since_2000)
    vondrak_dec = np.rad2deg(vondrak_dec)
    logger.debug(f"Precession for Star = {star_name}, Declination = {vondrak_dec}, RA = {vondrak_ra}")
    return vondrak_dec, vondrak_ra

def _generate_stereographic_projection(starList=None, 
                                    pole=None, 
                                    year_since_2000=None,
                                    is_precession=None,
                                    max_magnitude=None,
                                    declination_min=None,
                                    declination_max=None):
    # Generate stereographic projections and return declination and right ascension

    # Convert Star chart from RA hours to Radians to chart
    list_of_stars = _ra_to_radians(starList)

    finalPositionOfStarsDict = {} # {'Common Name': {"Declination" : Declination (float), "RA": RA (str)}
    x_star_labels = []
    x_ra_values = []
    y_dec_values = []
    for star in list_of_stars:
        # [[star_name, star_ra, star_declination, star_pm_speed, star_pm_angle, star_mag]]
        name = star[0]
        ra = star[1]
        dec = star[2]
        pm_speed = float(star[3])
        pm_angle = float(star[4])
        mag = float(star[5])
        if max_magnitude is None or mag < max_magnitude: # Optional: Filter out stars with a magnitude greater than max_magnitude
            logger.debug(f"Star = '{name}'")

            radius_of_circle = star_chart_spherical_projection._calculate_radius_of_circle(declination_min, pole)

            # Calculate position of star due to PROPER MOTION (changes RA and Declination over time)
            logger.debug(f"'{name}' original RA = {np.rad2deg(ra)} and Declination = {dec}")
            star_ra, star_declination = _ra_dec_via_pm(year_since_2000, 
                                                        ra, 
                                                        dec, 
                                                        pm_speed, 
                                                        pm_angle)
            logger.debug(f"Adjusted: {ra} RA (radians) = {star_ra}")
            logger.debug(f"Adjusted via Proper Motion: '{ra}': {dec} Declination (degrees) = {star_declination} ")

            # Optional: Calculate new position of star due to PRECESSION (change RA and Declination over time)
            # Vondrak accurate up  +/- 200K years around 2000
            if is_precession:
                star_declination, star_ra = _precession_vondrak(name, star_ra, star_declination, year_since_2000)
                logger.debug(f"Precession: {star_ra} RA (radians)\nPrecession: Declination (degrees) = {star_declination}")

                # convert degree to position on radius
                dec_ruler_position = star_chart_spherical_projection._calculate_length(star_declination, radius_of_circle, pole) 

                logger.debug(f"{name}: {star_declination} declination = {dec_ruler_position:.4f} cm")

                in_range_value = False # Determine if within range of South/North Hemisphere
                if star_declination > declination_min and star_declination < declination_max: # only display stars within range of declination values
                    in_range_value = True # North
                if star_declination < declination_min and star_declination > declination_max: # only display stars within range of declination values
                    in_range_value = True # South

                if in_range_value:
                    finalPositionOfStarsDict[name] = {"Declination" : float(star_declination), "RA": _radians_to_ra(star_ra)} # {'Common Name': {"Declination" : Declination (int), "RA": RA (str)}
                    x_star_labels.append(name)
                    x_ra_values.append(star_ra)
                    y_dec_values.append(dec_ruler_position)
                    logger.debug(f"Original: '{name}': {np.rad2deg(ra)} RA (degrees) and {dec} Declination (degrees)")
            if not is_precession:
                dec_ruler_position = star_chart_spherical_projection._calculate_length(star_declination, radius_of_circle, pole) # convert degree to position on radius

                logger.debug(f"{name}: {star_declination} declination = {dec_ruler_position:.4f} cm")
                in_range_value = False # Determine if within range of South/North Hemisphere
                if star_declination > declination_min and star_declination < declination_max: # only display stars within range of declination values
                    in_range_value = True # North
                if star_declination < declination_min and star_declination > declination_max: # only display stars within range of declination values
                    in_range_value = True # South

                if in_range_value:
                    finalPositionOfStarsDict[name] = {"Declination" : float(star_declination), "RA": _radians_to_ra(star_ra)} # {'Common Name': {"Declination" : Declination (int), "RA": RA (str)}
                    x_star_labels.append(name)
                    x_ra_values.append(star_ra)
                    y_dec_values.append(dec_ruler_position)
                    logger.debug(f"Original: '{name}': {np.rad2deg(ra)} RA (degrees) and {dec} Declination (degrees)")

    return x_star_labels, x_ra_values, y_dec_values, finalPositionOfStarsDict

def plot_stereographic_projection(included_stars=[], 
                                pole=None, 
                                declination_min=None,
                                year_since_2000=0,
                                display_labels=True,
                                display_dec=True,
                                increment=10,
                                is_precession=True,
                                max_magnitude=None,
                                added_stars=[],
                                only_added_stars=False,
                                show_plot=True,
                                fig_plot_title=None,
                                fig_plot_color="C0",
                                figsize_n=12,
                                figsize_dpi=100,
                                save_plot_name=None):

    # Catch errors in given arguments before plotting and set default constants
    star_chart_spherical_projection.errorHandling(isPlotFunction=True,
                                                included_stars=included_stars,
                                                pole=pole, 
                                                declination_min=declination_min,
                                                year_since_2000=year_since_2000,
                                                display_labels=display_labels,
                                                display_dec=display_dec,
                                                increment=increment, 
                                                is_precession=is_precession,
                                                max_magnitude=max_magnitude,
                                                added_stars=added_stars,
                                                only_added_stars=only_added_stars,
                                                show_plot=show_plot,
                                                fig_plot_title=fig_plot_title,
                                                fig_plot_color=fig_plot_color,
                                                figsize_n=figsize_n,
                                                figsize_dpi=figsize_dpi,
                                                save_plot_name=save_plot_name)
    pole = pole.capitalize()
    listOfStars = []
    if not only_added_stars:
        included_stars = [x.title() for x in included_stars] # convert all names to capitalized
        for star in _get_stars(included_stars):
            listOfStars.append([star[0],
                                star[1],
                                star[2],
                                star[4],
                                star[5],
                                star[3]])
        for star_object in added_stars:
            star_row = [star_object.star_name,
                        star_object.ra,
                        star_object.dec,
                        star_object.pm_speed,
                        star_object.pm_angle,
                        star_object.magnitude]
            listOfStars.append(star_row)
    else:
        listOfStars = []
        for star_object in added_stars:
            star_row = [star_object.star_name,
                        star_object.ra,
                        star_object.dec,
                        star_object.pm_speed,
                        star_object.pm_angle,
                        star_object.magnitude]
            listOfStars.append(star_row)

    # plot star chart as a circular graph

    # Set declination based on hemisphere selected
    if declination_min is None:
        if pole == "North": declination_min = northern_declination_min
        if pole == "South": declination_min = southern_declination_min
    if pole == "North": declination_max = northern_declination_max
    if pole == "South": declination_max = southern_declination_max

    # Polar plot figure
    fig = plt.figure(figsize=(figsize_n,figsize_n), dpi=figsize_dpi)
    ax = fig.subplots(subplot_kw={'projection': 'polar'})

    # Set Declination (astronomical 'latitude') as Y (radius of polar plot)

    # Split up chart into North/South hemisphere
    declination_values = np.arange(declination_min, declination_max+1, increment) # +1 to show max value in range

    # Store the ruler positions based on degrees and the ratio of the ruler
    ruler_position_dict = star_chart_spherical_projection._calculate_ruler(declination_min,
                                                                        declination_max,
                                                                        increment, 
                                                                        pole)

    # Display declination lines on the chart from -min to +max
    def displayDeclinationMarksOnAxis(declination_values, dec_min, dec_max, isInverted):
        # set declination marks based on the ruler to space out lines
        ruler_declination_position = list(ruler_position_dict.values())
        ruler_declination_labels = list(ruler_position_dict.keys())
        both_label_values = [list(x) for x in zip(ruler_declination_position, ruler_declination_labels)] # for testing
        ax.set_ylim(0, max(ruler_declination_position))

        # Display Axis
        if display_dec:
            ruler_declination_labels = [f"{deg}°" for deg in ruler_declination_labels]
            plt.yticks(ruler_declination_position, fontsize=7)
            ax.set_yticklabels(ruler_declination_labels)
            ax.set_rlabel_position(120) # declination labels position
        else:
            plt.yticks(ruler_declination_position, fontsize=0) # do not display axis
            ax.set_yticklabels(ruler_declination_labels)
            ax.set_rlabel_position(120) # declination labels position

    # Display declination lines based on hemisphere
    if pole == "North":
        displayDeclinationMarksOnAxis(declination_values, northern_declination_min, northern_declination_max, False)
    if pole == "South":
        displayDeclinationMarksOnAxis(declination_values, southern_declination_min, southern_declination_max, True)

    logger.debug(f"\n{pole}ern Range of Declination: {declination_min} to {declination_max}")

    # convert to x and y values for stars
    x_star_labels, x_ra_values, y_dec_values, star_dict = _generate_stereographic_projection(starList=listOfStars, 
                                                                                        pole=pole, 
                                                                                        year_since_2000=year_since_2000,
                                                                                        is_precession=is_precession,
                                                                                        max_magnitude=max_magnitude,
                                                                                        declination_min=declination_min,
                                                                                        declination_max=declination_max)

    # Set Right Ascension (astronomical 'longitude') as X (theta of polar plot)
    angles_ra = np.array([0, 15, 30, 45, 60, 75, 90, 105, 120, 135, 150,
                        165, 180, 195, 210, 225, 240, 255, 270, 285, 300,
                        315, 330, 345])
    plt.xticks(angles_ra * np.pi / 180, fontsize=8)
    labels_ra = np.array(['$0^h$','$1^h$','$2^h$','$3^h$', '$4^h$','$5^h$',
                        '$6^h$','$7^h$', '$8^h$','$9^h$', '$10^h$',
                        '$11^h$','$12^h$','$13^h$','$14^h$','$15^h$',
                        '$16^h$','$17^h$','$18^h$','$19^h$','$20^h$', 
                        '$21^h$', '$22^h$','$23^h$'])
    ax.set_xticklabels(labels_ra, fontsize=10)

    # Optional: Label the stars with names
    if display_labels:
        for i, txt in enumerate(x_star_labels):
            ax.annotate(txt, (x_ra_values[i], y_dec_values[i]), 
                        horizontalalignment='center', verticalalignment='bottom', 
                        fontsize=8)
    for i, txt in enumerate(x_star_labels):
        logger.debug(f"{txt}: {np.rad2deg(x_ra_values[i]):05f} RA (degrees) and {y_dec_values[i]:05f} Declination (ruler)")
        output_string = "Proper Motion"
        logger.debug(f"{output_string} for {year_since_2000} Years\n")

    ax.scatter(x_ra_values, y_dec_values, s=10, c=fig_plot_color)

    # Set Default Figure Title based on variables used in calculation
    years_for_title = year_since_2000
    suffix = ""
    if 1000 <  abs(years_for_title) and abs(years_for_title) < 1000000:
        years_for_title = years_for_title / 1000
        suffix = "K"
    if abs(years_for_title) > 1000000:
        years_for_title = years_for_title / 1000000
        suffix = "M"
    if year_since_2000 >= -2000: year_bce_ce = f"{year_since_2000 + 2000} C.E" # positive years for C.E
    if year_since_2000 < -2000: year_bce_ce = f"{abs(year_since_2000 + 2000)} B.C.E" # negative years for B.C.E
    figure_has_precession_extra_string = "with Precession" if is_precession else "without Precession"

    if fig_plot_title is None: # by default sets title of plot
        ax.set_title(f"{pole}ern Hemisphere [{years_for_title}{suffix} Years Since 2000 ({year_bce_ce})]: {declination_max}° to {declination_min}° {figure_has_precession_extra_string}")
    else:
        ax.set_title(fig_plot_title)

    # Optional: Save plot with user-defined name/location
    if save_plot_name is not None: 
        fig.savefig(save_plot_name)

    # Optional: Show the plot when it has been calculated
    if show_plot:
        plt.show()
    else:
        plt.close()



# -*- coding: utf-8 -*-
########################################################################
# Python Implementation of "New precession expressions, valid for long time intervals" (J. Vondrak, N. Capitaine, and P. Wallace) (2011)
# Code from: Dreamalligator - Vondrak (set for python3 here)
# https://github.com/dreamalligator/vondrak/blob/master/vondrak/__init__.py
########################################################################
from numpy import array, sin, cos, sqrt, append
import math

# 2Pi
TAU = 6.283185307179586476925287e0

# Arcseconds to radians
AS2R = 4.848136811095359935899141e-6

# Obliquity at J2000.0 (radians)
EPS0 = 84381.406 * AS2R

def ltp_pecl(jepoch):
    '''Long-term precession of the ecliptic
    Given Julian epoch (TT).
    Return ecliptic pole unit vector.
    ``ltp_pecl`` generates the unit vector for the pole of the ecliptic, using
    the series for Pₐ, Qₐ. The vector is with respect to the J2000.0 mean
    equator and equinox.
    '''

    # There is a typographical error in the original coefficient
    # C₇ for Qₐ should read 198.296701 instead of 198.296071
    # Src:
    # http://www.aanda.org/articles/aa/abs/2012/05/aa17274e-11/aa17274e-11.html

    # Number of polynomial and periodic coefficients
    npol = 4
    nper = 8

    # Polynomial coefficients
    pqpol = array([
        [+5851.607687, -1600.886300],
        [-0.1189000, +1.1689818],
        [-0.00028913, -0.00000020],
        [+0.000000101, -0.000000437]
    ])

    # Periodic coefficients
    pqper = array([
        [708.15, -5486.751211, -684.661560, 667.666730, -5523.863691],
        [2309.00, -17.127623, 2446.283880, -2354.886252, -549.747450],
        [1620.00, -617.517403, 399.671049, -428.152441, -310.998056],
        [492.20, 413.442940, -356.652376, 376.202861, 421.535876],
        [1183.00, 78.614193, -186.387003, 184.778874, -36.776172],
        [622.00, -180.732815, -316.800070, 335.321713, -145.278396],
        # corrected coefficient per erratum
        [882.00, -87.676083, 198.296701, -185.138669, -34.744450],
        [547.00, 46.140315, 101.135679, -120.972830, 22.885731]
    ])

    # Centuries since J2000
    T = (jepoch-2000.0)/100.0

    # Initialize Pₐ and Qₐ accumulators
    P = 0.0
    Q = 0.0

    # Periodic Terms
    for i in range(0, nper):
        W = TAU*T
        A = W/pqper[i][0]
        S = sin(A)
        C = cos(A)
        P = P + C*pqper[i][1] + S*pqper[i][3]
        Q = Q + C*pqper[i][2] + S*pqper[i][4]

    # Polynomial Terms
    W = 1.0
    for i in range(0, npol):
        P = P + pqpol[i][0]*W
        Q = Q + pqpol[i][1]*W
        W = W*T

    # Pₐ and Qₐ (radians)
    P = P*AS2R
    Q = Q*AS2R

    # Form the ecliptic pole vector
    Z = sqrt(max(1.0-P*P-Q*Q, 0.0))
    S = sin(EPS0)
    C = cos(EPS0)
    vec0 = P
    vec1 = -Q*C - Z*S
    vec2 = -Q*S + Z*C
    vec = array([vec0, vec1, vec2])
    return vec


def ltp_pequ(jepoch):
    '''Long-term precession of the equator
    Given Julian epoch (TT)
    Return equator pole unit vector
    ``ltp_pequ`` generates the unit vector for the pole of the equator, using
    the series for Xₐ, Yₐ. The vector is with respect to the J2000.0 mean
    equator and equinox.
    '''

    # Number of polynomial and periodic coefficients
    npol = 4
    nper = 14

    # Polynomial coefficients
    xypol = array([
        [+5453.282155, -73750.930350],
        [+0.4252841, -0.7675452],
        [-0.00037173, -0.00018725],
        [-0.000000152, +0.000000231]
    ])

    # Periodic coefficients
    xyper = array([
        [256.75, -819.940624, 75004.344875, 81491.287984, 1558.515853],
        [708.15, -8444.676815, 624.033993, 787.163481, 7774.939698],
        [274.20, 2600.009459, 1251.136893, 1251.296102, -2219.534038],
        [241.45, 2755.175630, -1102.212834, -1257.950837, -2523.969396],
        [2309.00, -167.659835, -2660.664980, -2966.799730, 247.850422],
        [492.20, 871.855056, 699.291817, 639.744522, -846.485643],
        [396.10, 44.769698, 153.167220, 131.600209, -1393.124055],
        [288.90, -512.313065, -950.865637, -445.040117, 368.526116],
        [231.10, -819.415595, 499.754645, 584.522874, 749.045012],
        [1610.00, -538.071099, -145.188210, -89.756563, 444.704518],
        [620.00, -189.793622, 558.116553, 524.429630, 235.934465],
        [157.87, -402.922932, -23.923029, -13.549067, 374.049623],
        [220.30, 179.516345, -165.405086, -210.157124, -171.330180],
        [1200.00, -9.814756, 9.344131, -44.919798, -22.899655]
    ])

    # Centuries since J2000
    T = (jepoch-2000.0)/100.0

    # Initialize Pₐ and Qₐ accumulators
    X = 0.0
    Y = 0.0

    # Periodic Terms
    for i in range(0, nper):
        W = TAU*T
        A = W/xyper[i][0]
        S = sin(A)
        C = cos(A)
        X = X + C*xyper[i][1] + S*xyper[i][3]
        Y = Y + C*xyper[i][2] + S*xyper[i][4]

    # Polynomial Terms
    W = 1.0
    for i in range(0, npol):
        X = X + xypol[i][0]*W
        Y = Y + xypol[i][1]*W
        W = W*T

    # X and Y (direction cosines)
    X = X*AS2R
    Y = Y*AS2R

    # Form the equator pole vector
    veq0 = X
    veq1 = Y
    W = X*X + Y*Y
    veq2 = 0.0
    if(W < 1.0):
        veq2 = sqrt(1.0-W)
    veq = array([veq0, veq1, veq2])
    return veq

def pxp(a, b):
    '''p-vector outer (=vector=cross) product.
    Given two p-vectors (a and b)
    Return a x b
    '''

    xa = a[0]
    ya = a[1]
    za = a[2]
    xb = b[0]
    yb = b[1]
    zb = b[2]
    axb = array([])
    axb = append(axb, ya*zb - za*yb)
    axb = append(axb, za*xb - xa*zb)
    axb = append(axb, xa*yb - ya*xb)
    return axb

def pn(p):
    '''Convert a p-vector into modulus and unit vector.
    Given p-vector (p)
    Return modulus (r), and unit vector (u)
    '''

    # Modulus of p-vector
    # http://www.iausofa.org/2013_1202_C/sofa/pm.c
    w = sqrt(p[0]*p[0] + p[1]*p[1] + p[2]*p[2])
    u = array([])
    if(w == 0.0):
        # zero a p-vector
        # http://www.iausofa.org/2013_1202_C/sofa/zp.c
        u = append(u, [0.0, 0.0, 0.0])
    else:
        # unit vector
        # http://www.iausofa.org/2013_1202_C/sofa/sxp.c
        s = 1.0/w
        u = append(u, s*p[0])
        u = append(u, s*p[1])
        u = append(u, s*p[2])
    r = w
    return r, u


def pdp(a, b):
    '''p-vector inner (=scalar=dot) product.
    Given two p-vectors (a and b)
    Return a * b
    '''

    # this is akin to the SOFA iauPdp, but it handles 3x3 square
    # matrices times (dot) column vectors 3x1
    # http://www.iausofa.org/2013_1202_C/sofa/pdp.c
    row = []
    for r in range(3):
        w = a[r][0]*b[0] + a[r][1]*b[1] + a[r][2]*b[2]
        row.append(w)
    adb = array(row)
    return adb


def ltp_pmat(jepoch):
    '''Long-term precession matrix
    Given Julian epoch (TT)
    Return precession matrix, J2000.0 to date
    ``ltp_pmat`` generates the 3 × 3 rotation matrix **P**, constructed using
    Fabri parameterization (i.e. directly from the unit vectors for the
    ecliptic and equator poles). The resulting matrix transforms vectors with
    respect to the mean equator and equinox of epoch 2000.0 into mean place of
    date.
    The matrix is in the sense ``P_date = RP x P_J2000``, where ``P_J2000`` is
    a vector with respect to the J2000.0 mean equator and equinox and
    ``P_date`` is the same vector with respect to the equator and equinox of
    epoch EPJ.
    '''

    # Equator pole (bottom row of matrix)
    peqr = ltp_pequ(jepoch)

    # Ecliptic pole
    pecl = ltp_pecl(jepoch)

    # Equinox (top row of matrix)
    V = pxp(peqr, pecl)  # P-vector outer product.
    w, EQX = pn(V)  # Convert a p-vector into modulus and unit vector

    # Middle row of matrix
    V = pxp(peqr, EQX)

    # The matrix elements
    rp = array([])
    rp = append(rp, [EQX, V, peqr])
    rp = rp.reshape(3, 3)
    return rp


def ltp_pbmat(jepoch):
    '''Long-term precession matrix, including GCRS frame bias.
    Given Julian epoch (TT)
    Return precession-bias matrix, J2000.0 to date
    ``ltp_pbmat`` generates the 3 × 3 rotation matrix **P** × **B**, where
    **B** is the “frame bias matrix” that expresses the relative orientation of
    the GCRS and mean J2000.0 reference systems. A simple first-order
    implementation of the frame bias is used, the departure from rigor being
    well under 1 μas.
    The matrix is in the sense ``P_date = RPB x P_J2000``, where ``P_J2000`` is
    a vector in the Geocentric Celestial Reference System, and P_date is the
    vector with respect to the Celestial Intermediate Reference System at that
    date but with nutation neglected.
    A first order bias formulation is used, of sub-microarcsecond accuracy
    compared with a full 3D rotation.
    '''

    # Frame bias (IERS Conventions 2010, Eqs. 5.21 and 5.33)
    DX = -0.016617 * AS2R
    DE = -0.0068192 * AS2R
    DR = -0.0146 * AS2R

    # Precession matrix.
    rp = ltp_pmat(jepoch)

    # Apply the bias
    rpb = array([])
    rpb = append(rpb, rp[0][0] - rp[0][1]*DR + rp[0][2]*DX)
    rpb = append(rpb, rp[0][0]*DR + rp[0][1] + rp[0][2]*DE)
    rpb = append(rpb, -rp[0][0]*DX - rp[0][1]*DE + rp[0][2])
    rpb = append(rpb, rp[1][0] - rp[1][1]*DR + rp[1][2]*DX)
    rpb = append(rpb, rp[1][0]*DR + rp[1][1] + rp[1][2]*DE)
    rpb = append(rpb, -rp[1][0]*DX - rp[1][1]*DE + rp[1][2])
    rpb = append(rpb, rp[2][0] - rp[2][1]*DR + rp[2][2]*DX)
    rpb = append(rpb, rp[2][0]*DR + rp[2][1] + rp[2][2]*DE)
    rpb = append(rpb, -rp[2][0]*DX - rp[2][1]*DE + rp[2][2])
    rpb = rpb.reshape(3, 3)
    return rpb


def epj(dj):
    '''Julian Date to Julian Epoch'''
    # based on http://www.iausofa.org/2013_1202_C/sofa/epj.c
    DJ00 = 2451545.0  # Reference epoch (J2000.0), Julian Date
    DJY = 365.25  # Days per Julian year
    jepoch = 2000.0 + (dj - DJ00)/DJY
    return jepoch


def ra_dec(v):
    '''Convert a cartesian position matrix to RA and Dec'''
    x = v[0][0]
    y = v[1][0]
    z = v[2][0]
    ra = math.atan2(y, x)
    dec = math.atan2(z, sqrt(x*x + y*y))
    return ra, dec



import math
import numpy as np

import star_chart_spherical_projection

class add_new_star:
    def __init__(self,
                star_name=None,
                ra=None,
                dec=None,
                pm_speed=None,
                pm_angle=None,
                pm_speed_ra=None,
                pm_speed_dec=None,
                magnitude=None):

        star_chart_spherical_projection.errorHandlingStarClass(star_name=star_name,
                                                                ra=ra,
                                                                dec=dec,
                                                                pm_speed=pm_speed,
                                                                pm_angle=pm_angle,
                                                                pm_speed_ra=pm_speed_ra,
                                                                pm_speed_dec=pm_speed_dec,
                                                                magnitude=magnitude)

        self.star_name = star_name
        self.ra = ra
        self.dec = dec
        self.magnitude = magnitude
        self.pm_speed_ra = pm_speed_ra
        self.pm_speed_dec = pm_speed_dec

        if self.pm_speed_ra is not None and self.pm_speed_dec is not None:
            self.convertToSpeedAndAngle(proper_motion_ra=self.pm_speed_ra, proper_motion_dec=self.pm_speed_dec)
        else:
            self.pm_speed = pm_speed
            self.pm_angle = pm_angle

    def convertToSpeedAndAngle(self, proper_motion_ra=None, proper_motion_dec=None):
        # convert proper motion ra and declination to a speed and angle
        self.pm_speed = math.sqrt(proper_motion_dec**2 + proper_motion_ra**2)
        self.pm_angle = np.rad2deg(math.atan(proper_motion_ra / proper_motion_dec))
        if self.pm_angle < 0: self.pm_angle += 360 # clamp between 0 and 360



########################################################################
# Distance between declination values in the radius are defined as either
#       Northern Hemisphere = tan(45° - angle of inclination)
#       Southern Hemisphere = tan(45° + angle of inclination)
########################################################################
import math
import numpy as np

def _calculate_length(angle_of_inclination, radius_of_circle, pole):
    # convert angle into length of radius
    if pole == "North":
        angle_in_radians = np.deg2rad(45 - angle_of_inclination/2) # + angle for northern projection
    if pole == "South":
        angle_in_radians = np.deg2rad(45 +  angle_of_inclination/2) # - angle for southern projection
    equation_of_length = radius_of_circle * math.tan(angle_in_radians) # calculated
    return equation_of_length

def _calculate_radius_of_circle(min_dec, pole):
    # calculate radius of full circle from -80 to 80 where min dec is the radius of smaller circle
    # assumes total length = 1
    total_ruler_len = 1
    if pole == "North":
        radius_of_circle_at_min_dec = (total_ruler_len/2) / math.tan(np.deg2rad(45 - min_dec/2))
    if pole == "South":
        radius_of_circle_at_min_dec = (total_ruler_len/2) / math.tan(np.deg2rad(45 + min_dec/2))
    return radius_of_circle_at_min_dec

def _calculate_ruler(declination_min, declination_max, increment, pole):
    # define the length of each segment in ruler when radius = 1

    x_angleOfDeclination = np.arange(-90, 90+1,increment) # declination max range from -90 to 90
    y_lengthSegments = []

    declination_angles_ruler = np.arange(-90, 90+1, increment) # declination max range from -90 to 90

    # calculate full size of circle to find declination for smaller range
    radius_of_circle = _calculate_radius_of_circle(declination_min, pole)

    ruler_position_dict = {} # dict: {degree : position_on_ruler }

    for n_angle in declination_angles_ruler:
        if pole == "North":
            ruler_position = _calculate_length(n_angle, radius_of_circle, "North")
            y_lengthSegments.append(ruler_position)
            if n_angle >= declination_min and n_angle <= declination_max: # North
                ruler_position_dict[n_angle] = round(ruler_position, 4)
        if pole == "South":
            ruler_position = _calculate_length(n_angle, radius_of_circle, "South")
            y_lengthSegments.append(ruler_position)
            if n_angle <= declination_min and n_angle >= declination_max: # South
                ruler_position_dict[n_angle] = round(ruler_position, 4)

    return ruler_position_dict



# generate_star_chart.py function calls
from .generate_star_chart import plot_stereographic_projection
from .generate_star_chart import _get_stars
from .generate_star_chart import _ra_to_radians
from .generate_star_chart import _radians_to_ra
from .generate_star_chart import _generate_stereographic_projection
from .generate_star_chart import _ra_dec_via_pm
from .generate_star_chart import _precession_vondrak

# position_of_stars.py function calls
from .position_of_stars import final_position
from .position_of_stars import position_over_time
from .position_of_stars import plot_position
from .position_of_stars import predict_pole_star

# starClass.py function calls
from .starClass import add_new_star

# declination_r_axis.py function calls
from .declination_r_axis import _calculate_ruler
from .declination_r_axis import _calculate_radius_of_circle
from .declination_r_axis import _calculate_length

## Error Handling: error_handling.py function calls
from .error_handling import errorHandling
from .error_handling import errorHandlingStarClass
from .error_handling import errorHandlingPredictPoleStar

## TEMP: Vondrak in Python 3
from .ra_dec_precession_vondrak import ltp_pbmat
from .ra_dec_precession_vondrak import pdp
from .ra_dec_precession_vondrak import ra_dec
